/*
 * Subverse
 * Copyright (C) 2009 Saulo Popov Zambiasi
 *
 * Subverse is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Subverse is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: saulopz@gmail.com
 */

#include "World.h"

using namespace std;

World::World(std::string fileName) {
	background = create_bitmap(MAX_X + (CELL_SIZE * 2), MAX_Y + (CELL_SIZE * 2));
	worldScreen = create_bitmap(MAX_X, MAX_Y);
	timer = new Timer();
	height = 0;
	width = 0;
	posX = 0;
	posY = 0;
	elementInfo = NULL;
	avatar = NULL;
	music = load_midi("clkcwhrv.mid");
	turnGame = 0;
	step = 0;
	play_midi(music, TRUE);
	indexId = 0;
	srand (time(NULL));

	images = new Images();
	images->add("meat");
	images->add("dead");
	images->add("avatar");
	images->add("cyclope");
	images->add("water");
	images->add("adobe");
	images->add("spider");
	images->add("ground");
	images->add("fire");
	images->add("lawn");
	images->add("littlelawn");
	images->add("rocks");
	images->add("tree");
	images->add("wall");

	loadWorld(fileName);
}

int World::getTurnGame() {
    return turnGame;
}

long int World::getId() {
    indexId++;
    return indexId;
}

Element* World::get(Element* e, int type, int direction) {
	Element* aux = NULL;
	switch(direction) {
        case DOWN_RIGHT :
            if ((e->getY() < getHeight()-1) && (e->getX() < getWidth()-1)) {
                aux = get(type, e->getX()+1, e->getY()+1);
            }
            break;
        case DOWN_LEFT:
            if ((e->getY() < getHeight()-1) && (e->getX() > 0)) {
                aux = get(type, e->getX()-1, e->getY()+1);
            }
            break;
        case UP_RIGHT :
            if ((e->getY() > 0) && (e->getX() < getWidth()-1)) {
                aux = get(type, e->getX()+1, e->getY()-1);
            }
            break;
        case UP_LEFT:
            if ((e->getY() > 0) && (e->getX() > 0)) {
                aux = get(type, e->getX()-1, e->getY()-1);
            }
            break;
		case UP   :
			if (e->getY() > 0) {
				aux = get(type, e->getX(), e->getY()-1);
			}
			break;
		case DOWN :
			if (e->getY() < getHeight()-1) {
				aux = get(type, e->getX(), e->getY()+1);
			}
			break;
		case LEFT :
			if (e->getX() > 0) {
				aux = get(type, e->getX()-1, e->getY());
			}
			break;
		case RIGHT:
			if (e->getX() < getWidth()-1) {
				aux = get(type, e->getX()+1, e->getY());
			}
			break;
	}
	return aux;
}

bool World::move(Element* e, int direction) {
	bool freeWay = false;
	Element* land = get(e, LAND, direction);
	if (land != NULL) {
		if (land->getType() < WALL) {
			freeWay = true;
		}
		Element *aux = get(e, ELEMENT, direction);
		if (aux != NULL) freeWay = false;
	}
	bool moved = false;
	if (freeWay) {
		int auxX = e->getX();
		int auxY = e->getY();
		switch(direction) {
            case DOWN_RIGHT : auxX++; auxY++; break;
            case DOWN_LEFT  : auxX--; auxY++; break;
            case UP_RIGHT   : auxX++; auxY--; break;
            case UP_LEFT    : auxX--; auxY--; break;
			case UP    : auxY--; break;
			case DOWN  : auxY++; break;
			case RIGHT : auxX++; break;
			case LEFT  : auxX--; break;
		}
		set(ELEMENT, auxX, auxY, e);
		set(ELEMENT, e->getX(), e->getY(), NULL);
		e->setX(auxX);
		e->setY(auxY);
		moved = true;
	}
	return moved;
}

void World::draw() {
	show_mouse(NULL);
	// draw land
	for (int x = (-1); x < getScreenWidth()+1; x++) {
		for (int y = (-1); y < getScreenHeight()+1; y++) {
			Element* e = get(LAND, getX() + x, getY() + y);
			if (e != NULL) {
				draw_sprite(background, e->getImage(), (x+1) * CELL_SIZE, (y+1) * CELL_SIZE);
			}
		}
	}
	// draw elements
	for (int x = (-1); x < getScreenWidth()+1; x++) {
		for (int y = (-1); y < getScreenHeight()+1; y++) {
			Element* e = get(ELEMENT, getX() + x, getY() + y);
			if (e != NULL) {
                int aux = (int) (CELL_SIZE / 2);
				int auxX = ((x+1) * CELL_SIZE) - aux;
				int auxY = ((y+1) * CELL_SIZE) - aux;
				if (e->isWalking() && e->isStep()) {
					switch (e->getDirection()) {
                        case DOWN_RIGHT : auxY -= aux; auxX -= aux; break;
                        case DOWN_LEFT  : auxY -= aux; auxX += aux; break;
                        case UP_RIGHT   : auxY += aux; auxX -= aux; break;
                        case UP_LEFT    : auxY += aux; auxX += aux; break;
						case DOWN  : auxY -= aux; break;
						case RIGHT : auxX -= aux; break;
						case UP    : auxY += aux; break;
						case LEFT  : auxX += aux; break;
					}
				}
                if (e == elementInfo) {
                    rect(background, auxX+25, auxY+25, auxX+75, auxY+75, makecol(200, 200, 0));
                }
                string talk = e->getTalk();
                if (talk.compare("") != 0) {
                    textout_centre_ex(background, font, talk.c_str(), auxX+CELL_SIZE, auxY, makecol(255,255,255), -1);
                }
                draw_sprite(background, e->getImage(), auxX, auxY);
			}
		}
	}
	int auxX = CELL_SIZE;
	int auxY = CELL_SIZE;
	int aux = (int) (CELL_SIZE / 2);
	// int aux = CELL_SIZE - (int) ((CELL_SIZE * avatar->getTime()) / avatar->getMaxTime());
	if (avatar != NULL) {
    	if (avatar->isWalking() && avatar->isStep()) {
    		switch (avatar->getDirection()) {
                case DOWN_RIGHT :
                    if (((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) &&
                        ((getX() > 0) && (getX() < (getWidth() - getScreenWidth())))) {
                        auxY -= aux;
                        auxX -= aux;
                    }
                    break;
                case DOWN_LEFT :
                    if (((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) &&
                        ((getX() > 0) && (getX() < (getWidth() - getScreenWidth())))) {
                        auxY -= aux;
                        auxX += aux;
                    }
                    break;
                case UP_RIGHT :
                    if (((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) &&
                        ((getX() > 0) && (getX() < (getWidth() - getScreenWidth())))) {
                        auxY += aux;
                        auxX -= aux;
                    }
                    break;
                case UP_LEFT :
                    if (((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) &&
                        ((getX() > 0) && (getX() < (getWidth() - getScreenWidth())))) {
                        auxY += aux;
                        auxX += aux;
                    }
                    break;
    			case UP    :
    				if ((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) {
    					auxY += aux;
    				}
    				break;
    			case DOWN  :
    				if ((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) {
    					auxY -= aux;
    				}
    				break;
    			case LEFT  :
    				if ((getX() > 0) && (getX() < (getWidth() - getScreenWidth()))) {
    					auxX += aux;
    				}
    				break;
    			case RIGHT :
    				if ((getX() > 0) && (getX() < (getWidth() - getScreenWidth()))) {
    					auxX -= aux;
    				}
    				break;
    		}
    	}
    }
    // Desenha outras informações
    blit(background, worldScreen, auxX, auxY, 0, 0, MAX_X, MAX_Y);
	if (elementInfo != NULL) {
		string strout = elementInfo->toString();
		textout_ex(worldScreen, font, strout.c_str(), 10, MAX_Y-15, makecol(255,255,255), -1);
	}
	show_mouse(worldScreen);
	blit(worldScreen, screen, 0, 0, 0, 0, MAX_X, MAX_Y);
}

void World::run() {
	Element *e = NULL;
	while (!key[KEY_ESC]) {
		for (int x = 0; x < getWidth(); x++) {
			for (int y = 0; y < getHeight(); y++) {
				e = get(ELEMENT, x, y);
				if (e != NULL) {
					e->run();
					if (e->isToErase()) {
						set(ELEMENT, x, y, NULL);
					}
				}
				e = get(LAND, x, y);
				if (e != NULL) {
					e->run();
					if (e->isToErase()) {
						set(LAND, x, y, NULL);
					}
				}
			}
		}
		setX(avatar->getX() - (int) (getScreenWidth() / 2));
		setY(avatar->getY() - (int) (getScreenHeight() / 2));
		incTimer();
    	if (mouse_b & 1) {
    		int mouseX = getX() + ((int) (mouse_x / CELL_SIZE));
    		int mouseY = getY() + ((int) (mouse_y / CELL_SIZE));
    		Element* eInfo = get(ELEMENT, mouseX, mouseY);
    		if (eInfo != NULL) {
    			elementInfo = eInfo;
    		}
    	}
		draw();
		clear_keybuf();
		rest(30);
	}
	if (step == 0) {
	    step = 1;
	} else {
	    step = 0;
	}
	turnGame++;
}

void World::loadWorld(std::string fileName) {
	LoadWorld *loadWorld = new LoadWorld(fileName);
	height = loadWorld->getHeight();
	width = loadWorld->getWidth();

	Type* typeLand = new Type();
	Type* typeElement = new Type();

	for (int y = 0; y < getHeight(); y++) {
		Line* lineElement = new Line();
		Line* lineLand = new Line();
    	for (int x = 0; x < getWidth(); x++) {
    		Element *e;
    		int iten = loadWorld->getNext(ELEMENT);
    		switch (iten) {
    			case AVATAR :
    				if (avatar == NULL)	{
    					e = new Avatar(this);
    					avatar = e;
    					setX(x - (int) (getScreenWidth() / 2));
    					setY(y - (int) (getScreenHeight() / 2));
    				} else e = NULL;
    				break;
    			case CYCLOPE : e = new Cyclope(this); break;
    			case SPIDER : e = new Spider(this); break;
    			case FIRE : e = new Element(this, "fire", FIRE); break;
    			case TREE : e = new Element(this, "tree", TREE); break;
    			case WALL : e = new Element(this, "wall", WALL); break;
    			default : e = NULL;
    		}
    		if (e != NULL) {
    			e->setX(x);
    			e->setY(y);
    			e->setTurnGame(turnGame);
    		}
    		lineElement->x.push_back(e);

    		iten = loadWorld->getNext(LAND);
    		switch (iten) {
    			case ADOBE : e = new Element(this, "adobe", ADOBE); break;
    			case GROUND : e = new Element(this, "ground", GROUND); break;
    			case ROCKS : e = new Element(this, "rocks", ROCKS); break;
    			case LAWN : e = new Element(this, "lawn", LAWN); break;
    			case LITTLELAWN : e = new Element(this, "littlelawn", LITTLELAWN); break;
    			case WATER : e = new Element(this, "water", WATER); break;
    			default : e = NULL;
    		}
    		if (e != NULL) {
    			e->setX(x);
    			e->setY(y);
    			e->setTurnGame(getTurnGame());
    		}
    		e->setMaxTime((rand() % 4) + 10);
    		lineLand->x.push_back(e);
    	}
    	typeElement->y.push_back(lineElement);
    	typeLand->y.push_back(lineLand);
    }
    element.push_back(typeLand);
    element.push_back(typeElement);
}

Vision* World::getVision(Element* e, int range) {
	Vision* vision = new Vision();

	if (e == NULL) return NULL;

	for (int i = 0; i <= range; i++) {
	    for (int j = -i-1; j <= i+1; j++) {
	        int auxX = e->getX();
	        int auxY = e->getY();
	        switch (e->getDirection()) {
                case UP_RIGHT  :
                case UP_LEFT   :
                case UP        : auxX += j; auxY -= i; break;
				case DOWN_RIGHT:
                case DOWN_LEFT :
                case DOWN      : auxX += j; auxY += i; break;
                case LEFT      : auxX -= i; auxY += j; break;
                case RIGHT     : auxX += i; auxY += j; break;
	        }
	        if ((auxX != getX()) && (auxY != getY())) {
                Element *element = get(ELEMENT, auxX, auxY);
                if (element != NULL) vision->add(element, ELEMENT);
                element = get(LAND, auxX, auxY);
                if (element != NULL) vision->add(element, LAND);
	        }
	    }
	}
	return vision;
}

void World::setMusic(bool m) {
    if (m) play_midi(music, TRUE);
    else stop_midi();
}

Element* World::get(int type, int x, int y) {
	if ((x < 0) || (x >= getWidth()) || (y < 0) || (y >= getHeight())) {
		return NULL;
	} else return element[type]->y[y]->x[x];
}

void World::set(int type, int x, int y, Element* e) {
	if ((x >= 0) && (x < getWidth()) &&
		(y >= 0) && (y < getHeight()) &&
		(type >= 0) && (type < 2)) {
		element[type]->y[y]->x[x] = e;
	}
}

int World::getWidth() {
	return width;
}

int World::getHeight() {
	return height;
}

int World::getX() {
	return posX;
}

int World::getY() {
	return posY;
}

int World::getScreenWidth() {
	return ((int) (MAX_X / CELL_SIZE));
}

int World::getScreenHeight() {
	return ((int) (MAX_Y / CELL_SIZE));
}

void World::setX(int in) {
	int temp = getScreenWidth();
	if (in > (width - temp)) {
		posX = width - temp;
	} else if (in < 0) {
		posX = 0;
	} else posX = in;
}

void World::setY(int in) {
	int temp = getScreenHeight();
	if (in > (height - temp)) {
		posY = height - temp;
	} else if (in < 0) {
		posY = 0;
	} else posY = in;
}

void World::setPosition(int x_in, int y_in) {
	setX(x_in);
	setY(y_in);
}

Timer *World::getTimer() {
	return timer;
}

BITMAP *World::getImage() {
	return worldScreen;
}

void World::incTimer() {
	timer->inc();
}

World::~World() {
	Element* e = NULL;
	for (int y = 0; y < getHeight(); y++) {
		for (int x = 0; x < getWidth(); x++) {
			e = get(ELEMENT, x, y);
			if (e != NULL) delete e;
			e = get(LAND, x, y);
			if (e != NULL) delete e;
		}
	}
	element.clear();
	delete images;
}
